@@ -40,10 +40,9 @@ pub enum Event {
 	},
 }
 
-impl Event {
-	pub(crate) fn emit(self) {
-		listener::with(|listener| listener.event(self));
-	}
+// Expose `listener::with` to the crate only.
+pub(crate) fn with<F: FnOnce(&mut (dyn EventListener + 'static))>(f: F) {
+	listener::with(f);
 }
 
 /// Run closure with provided listener.
